# 多线程数据流软件的优化

## 摘要

本文提出了一个优化工具，为多线程数据流软件找到最佳的线程数。假设线程封装并行可执行的关键功能，通过有限容量队列连接，并且需要某些硬件资源。我们展示了==基于排队理论==的测量和计算的组合如何导致递归地确定线程最佳组合的算法，即给定主机上多线程数据流软件的最佳配置。该算法在一个排队网络的有向图上进行，该网络对该软件进行了建模。

## 背景

cpu内部多核的趋势使软件工程朝着并发方向发展。软件被分割成可以并行运行的原子操作。这可能会大大加快计算速度，但也会因为线程协调而导致额外的开销

## 本文的方法

在这项工作中，我们使用术语主机来表示具有固定数量内核的服务器硬件平台。将节点指定为排队网络软件的任务，并构建数据流图。系统中的线程负责执行这些节点的任务。每个节点可以有多个并行执行节点任务的线程。总体情况是，我们希望找到每个节点的最佳线程数，并保证主机上可用的线程数量只能与内核数量相同，因为每个线程都在专用核心上运行。

==本文讨论了一种结合分析建模技术和测量的方法，通过迭代增加线程数量来找到特定的主机上线程的最佳配置==

## 相关工作

FFTW计划器通过测量许多不同计划的实际运行时间并选择最快的计划来工作

SPIRAL旨在自动生成用于线性数字信号处理(DSP)转换的高性能代码，并调整到给定平台。

通过猜测性能快速优化系统配置：开发了一种算法，用于寻找Web系统的近乎最佳配置，根据线程数和缓存大小，迭代地找到最优配置

GCD是用于在排队系统中组织任务的调度器，它的作用类似于线程管理器，对任务进行排队和调度，以便在处理器核心上并行执行。

用于软件架构性能预测的有限容量队列的排队网络模型综述：使用有限/单容量队列和服务后阻塞的排队网络模型来建模软件架构;更确切地说，同步约束和软件组件之间的同步通信。对组件之间的信息流(跟踪)进行分析，以确定通信类型(fork、join)，并揭示组件之间的交互对，从而能够对排队网络进行建模。这样就可以导出特定软件体系结构的性能模型

## 优化方法

我们的工作重点是为特定主机找到多线程数据流软件的最佳配置。一个配置被指定为包含n个元组的向量，`(k1 − ... − ki − ... − kn) `中ki为节点i的线程数,约束条件为`k1+k2+...+kn<=c`

其中c是特定主机上可用(虚拟)内核的数量，`(k1 − ... − ki − ... − kn)`初始化为1，优化是递归地通过向过度使用的节点添加线程来完成的，有两个优化目标:

- 吞吐量。通过对吞吐量进行优化，我们从每秒一个作业的最低到达率开始，然后工具逐步提高到达率，直到一个节点超过预定义的利用率阈值。然后，只要利用率低于阈值，相应的线程数量就会增加，同样，只要没有达到硬件限制，到达率就会提高，优化就会继续进行。对于最大可能的线程数，确定了最高可能的吞吐量

- 通过对整体进行优化，可以给出期望的到达率，并且该工具确定确保所有节点低于预定义的利用率阈值所需的最小线程数。

  在外部到达率恒定的情况下，只有利用率≥limU (0 < limU < 1)的过度使用节点的线程数才会增加，以便尽可能均匀地在可用内核之间分配负载。